/**
 * Return the formatted date of today.
 */
function get_formatted_date(split) {
    split = split || ".";
    var date = new Date();
    var tzoffset = (new Date()).getTimezoneOffset() * 60000;
    return ((new Date(Date.now() - tzoffset)).toISOString()).split(split)[0];
}

/**
 * This function retrieves an html form from the html subdirectory.
 *
 */
var getHTMLForm = async function (pageName, variables) {
    var site = await connection.get("webinterface/html/forms/" + pageName + ".html", "html");
    var htmltext = site.getElementById("caosdb-form").innerHTML;
    var req = /\{([0-9a-z-A-Z_]+)\}/g;
    return htmltext.replace(req, (match, p1) => {
        return variables[p1] || "";
    });
}

/**
 * GEOMAR loan management, copied and adapted from AWI box-management code.
 */
var loan_management = function () {

    var logger = log.getLogger("loan_management");

    var datamodel = {
        loan: "Loan",
        container: "Container",
        returned: "returned",
    }

    const _dismiss_button = '<button class="btn btn-secondary container-loan-btn">OK</button>'
    const _server_did_not_respond = "The server did not respond. Please reload the page.";
    const _return_container_button = '<a title="Return Container." class="btn btn-link container-loan-btn">Return Container</a>';
    const _borrow_container_button = '<a title="Borrow Container." class="btn btn-link container-loan-btn">Borrow Container</a>';
    const _confirm_loan_button = '<a title="Confirm Loan." class="btn btn-link container-loan-btn">Confirm Loan</a>';
    const _manual_return_button = '<a title="Manual Return." class="btn btn-link container-loan-btn">Manual Return</a>';
    const _accept_return_button = '<a title="Accept Return" class="btn btn-link container-loan-btn">Accept Return</a>';
    const _reject_return_button = '<a title="Reject Return." class="btn btn-link container-loan-btn">Reject Return</a>';
    const _accept_loan_button = '<a title="Accept Borrow Request." class="btn btn-link container-loan-btn">Accept Loan Request</a>';
    const _container_update_alert = `
        <div class="alert alert-danger alert-dismissible" role="alert">
            <h4>Update the Container</h4>
            <p>
                Are you sure that the container record is up-to-date (references etc.)?<br>Possibly have a look at the notes in the loan by clicking on the "References" button.<br> Do you  you want to apply further changes?
            </p><p>
            <button type="button" class="btn btn-danger caosdb-f-manual-return-cancel">Cancel and change the container manually</button>
            <button type="button" class="btn btn-secondary caosdb-f-manual-return-proceed">The container is up-to-date.</button>
        </p>
        </div>
    `;

    /**
     * Return a HTML span which tells the user to wait for something.
     *
     * The optional text parameter can be used to set the text that is shown.
     *
     * @param text
     * @return HTMLElement
     */
    var getPleaseWaitSpan = function (text = "Processing. Please wait...") {
        return $('<span>' + text + '</span>')[0];
    }

    /**
     * Return the actions panel of a container entity.
     */
    var get_actions_panel = function (container) {
        return $(container).find('.caosdb-entity-actions-panel');
    }

    /**
     * Open a mail with recepient address address, subject subject and body text.
     */
    this.open_mail_generic = function (address, subject, text) {
        const mailto = 'mailto:' + address + '?subject=' + subject + '&body=' + text;
        location.href = mailto;
    }

    /**
     * This function checks whether a property with name name
     * is present in the property list p.
     * This function is used for determining the loan state.
     */
    var loan_state_function = function (p, name) {
        var p1 = p.filter(x => x.name == name);
        if (p1.length == 1 && p1[0].value.length > 0) {
            return true;
        }
        return false;
    }

    /**
     * This function returns a string on the state of the loan.
     */
    var get_loan_state_string = function (loan) {
        const p = getProperties(loan);
        if (loan_state_function(p, datamodel.returned)) {
            return "returned";
        }
        if (loan_state_function(p, "returnAccepted")) {
            return "return_accepted";
        }
        if (loan_state_function(p, "returnRequested")) {
            return "return_requested";
        }
        if (loan_state_function(p, "lent")) {
            return "lent";
        }
        if (loan_state_function(p, "loanAccepted")) {
            return "loan_accepted";
        }
        if (loan_state_function(p, "loanRequested")) {
            return "loan_requested";
        }

        const loan_id = getEntityID(loan);
        throw new Error(`Unknown loan state for Loan ${loan_id}.`);
    }


    /**
     * Query for a Loan entity which references the container and which has no
     * `returned` property.
     */
    var get_active_loans = async function (container) {
        return await query(`FIND ${datamodel.loan} WITH ${datamodel.container} -> ${container.id} AND WHICH DOES NOT HAVE A ${datamodel.returned}`);
    }

    /**
     * Returns detailed information about the loan state of this container:
     */
    var get_loan_state = async function (container) {
        const loan_state = {
            loan: undefined,
            state: undefined,
        };
        const loan = (await get_active_loans(container))[0];
        if (typeof loan === "undefined") {
            // no loan found
            return loan_state;
        }

        loan_state.loan = loan
        loan_state.state = get_loan_state_string(loan);

        return loan_state;
    }

    var run_script = async function (script, data) {
        const json_str = JSON.stringify(data);
        const params = {
            "-p0": {
                "filename": "form.json",
                "blob": new Blob([json_str], {
                    type: "application/json"
                })
            }
        };
        try {
            const response = await connection.runScript(script, params);
            return form_elements.parse_script_result(response);
        } catch (error) {
            return error;
        }
    }

    var show_result = function (container, result, container) {
        const dismissable = $('<div/>');
        const restore = $(container).children();
        restore.hide();

        $(container).append(dismissable);

        if (typeof result === "undefined") {
            $(dismissable).append(_server_did_not_respond);
        }
        if (result["code"] == 0) {
            // success!
            $(dismissable).append(result["stdout"]);
        } else if (result["code"] > 0) {
            // :(
            $(dismissable)
                .append(result["stderr"])
                .append(result["stdout"]);
        } else {
            $(dismissable).append(result);
        }

        const dismiss_btn = $(_dismiss_button);
        dismiss_btn.click(function () {
            dismissable.remove();
            restore.show();
            init(container);
        });
        dismissable.append(dismiss_btn);
    }


    var get_request_data = function (form) {
        const email = $(form).find("#email").val();
        const first_name = $(form).find("#first-name").val();
        const last_name = $(form).find("#last-name").val();
        const comment = $(form).find("#loan-comment").val();
        const expected_return_date = $(form).find("#expected-return").val();
        const exhaust_contents = $(form).find("#exhaust-contents").prop("checked");

        window.localStorage["borrower_email"] = email;
        window.localStorage["borrower_first_name"] = first_name;
        window.localStorage["borrower_last_name"] = last_name;
        window.localStorage["borrower_return_date"] = expected_return_date

        return {
            email: email,
            first_name: first_name,
            last_name: last_name,
            comment: comment,
            expected_return_date: expected_return_date,
            exhaust_contents: exhaust_contents,
        };
    }


    /**
     * Generate a borrow function for a container.
     * This function can also be used to assign values to multiple containers.
     */
    var borrow_function = async function (form, container) {
        const wait = getPleaseWaitSpan();
        const actions_panel = get_actions_panel(container);
        $(actions_panel).append(wait).find('btn').hide()
        $(form).hide()

        const loan_request = get_request_data(form);


        $.extend(loan_request, {
            container: getEntityID(container),
        });

        const result = await run_script("loan_management/request_loan.py", loan_request);
        $(wait).remove();
        $(form).remove();
        show_result(actions_panel, result, container);
    }

    var _add_form = async function (container, form_generator, submit) {
        const form = $(await form_generator(container));
        form.insertAfter($(container).find('.caosdb-entity-actions-panel')).hide();
        form.submit(() => {
            submit(form[0], container);
            return false;
        });
        return form[0];
    }

    /**
     * Add Return button to containers' actions panels.
     */
    var add_return_button = async function (container) {
        const but = $(_return_container_button);

        get_actions_panel(container).append(but);
        const form = await _add_form(container, generate_return_form, return_function);
        $(form).find("label").css("display", "block");


        but.click(() => {
            but.remove();
            $(form).toggle()
        });
    }

    /**
     * If the user is administrator (or a data curator) a button is added
     * to accept the borrow request. Else a message is added "pending loan request".
     */
    var add_borrow_accept_button = function (container, loan) {
        const actionPanel = get_actions_panel(container);
        if (userIsAdministrator() === true || userHasRole("curator") === true) {
            const but = $(_accept_loan_button);

            but.click(() => {
                but.remove();
                accept_loan_function(container);
            });
            actionPanel.append(but);
        } else {
            actionPanel.append($(create_link_to_loan(loan, "Pending loan request.")));
        }
    }

    var create_link_to_loan = function (loan, text) {
        const loan_id = getEntityID(loan);
        const href = `${connection.getBasePath()}Entity/${loan_id}`;
        const title = "Open loan entity.";
        const classes = "btn btn-link container-loan-btn";
        return `<a title="${title}" class="${classes}"
                href="${href}">${text}<a>`;
    }

    /**
     * Add Borrow buttons to the containers' action panels
     */
    var add_borrow_button = async function (container) {
        const but = $(_borrow_container_button);
        get_actions_panel(container).append(but);

        const form = await _add_form(container, generate_form_borrow_checkout, borrow_function);
        $(form).find("label").css("display", "block");

        but.click(() => {
            but.remove();
            $(form).toggle()
        });
    }


    var _get_label_or_id = function (container) {
        const bn = getProperty(container, "recommendedLabel", case_sensitive = false);
        if (bn === undefined) {
            // ID as fallback
            return getEntityID(container);
        }
        return bn;

    }

    var open_accept_loan_mail = async function (loan, container) {
        try {
            const borrower = (await retrieve(getProperty(loan, "borrower"), case_sensitive = false))[0];

            const email = getProperty(borrower, "email", case_sensitive = false);
            const firstName = getProperty(borrower, "firstName", case_sensitive = false);
            const lastName = getProperty(borrower, "lastName", case_sensitive = false);
            const bn = _get_label_or_id(container);
            const date = getProperty(loan, "expectedReturn", case_sensitive = false);

            const subject = `Loan Request for Container ${bn} accepted`;
            const body = `Dear ${firstName} ${lastName}, %0D%0D` +
                `You can pickup container ${bn} on ${date}. %0D%0D` +
                `Kind Regards,%0D%0D ${getUserName()}`;
            open_mail_generic(email, subject, body)
        } catch (err) {
            logger.error(err);
        }
    }


    var accept_loan_function = async function (container) {

        const wait = getPleaseWaitSpan();
        const actions_panel = get_actions_panel(container);
        $(actions_panel).append(wait).find('btn').hide()

        const loan = (await get_active_loans(container))[0];

        const accept_loan_request = {
            loan: getEntityID(loan),
        };

        const result = await run_script("loan_management/accept_loan_request.py", accept_loan_request);
        $(wait).remove();
        show_result(actions_panel, result, container);

        open_accept_loan_mail(loan, container);

    }

    var manual_return_function = async function (container) {
        const wait = getPleaseWaitSpan();
        const actions_panel = get_actions_panel(container);
        $(actions_panel).append(wait).find('btn').hide()

        const loan = (await get_active_loans(container))[0];

        const manual_return_request = {
            loan: getEntityID(loan),
        };

        const result = await run_script("loan_management/manual_return.py", manual_return_request);
        $(wait).remove();
        show_result(actions_panel, result, container);
    }

    /**
     * Add Manual Return button to containers' actions panels.
     */
    var add_manual_return_button = function (container, loan) {
        const actionPanel = get_actions_panel(container);
        if (userIsAdministrator() === true || userHasRole("curator") === true) {
            const but = $(_manual_return_button);
            actionPanel.append(but);

            const container_update_alert = $(_container_update_alert);
            container_update_alert.find(".caosdb-f-manual-return-cancel").click(() => {
                but.show();
                container_update_alert.hide();
            });

            container_update_alert.find(".caosdb-f-manual-return-proceed").click(() => {
                but.remove();
                container_update_alert.remove();
                manual_return_function(container);
            });

            actionPanel.append(container_update_alert);
            container_update_alert.hide();

            but.click(() => {
                but.hide();
                container_update_alert.show();
            });
        } else {
            actionPanel.append($(create_link_to_loan(loan, "Pending return.")));
        }
    }


    var confirm_loan_function = async function (container) {
        const wait = getPleaseWaitSpan();
        const actions_panel = get_actions_panel(container);
        $(actions_panel).append(wait).find('btn').hide()

        const loan = (await get_active_loans(container))[0];

        const confirm_loan_request = {
            loan: getEntityID(loan),
        };

        const result = await run_script("loan_management/confirm_loan.py", confirm_loan_request);
        $(wait).remove();
        show_result(actions_panel, result, container);
    }


    var open_reject_return_mail = async function (loan, container) {
        try {
            const borrower = (await retrieve(getProperty(loan, "borrower"), case_sensitive = false))[0];

            const email = getProperty(borrower, "email", case_sensitive = false);
            const firstName = getProperty(borrower, "firstName", case_sensitive = false);
            const lastName = getProperty(borrower, "lastName", case_sensitive = false);
            const bn = _get_label_or_id(container);
            const subject = `Return Request for Container ${bn} rejected`;
            const body = `Dear ${firstName} ${lastName}, %0D%0D` +
                `your return request for container ${bn} cannot be accepted.` +
                `Kind Regards,%0D ${getUserName()}`;
            open_mail_generic(email, subject, body)
        } catch (err) {
            logger.error(err);
        }
    }


    var open_accept_return_mail = async function (loan, container) {
        try {
            const borrower = (await retrieve(getProperty(loan, "borrower"), case_sensitive = false))[0];

            const email = getProperty(borrower, "email", case_sensitive = false);
            const firstName = getProperty(borrower, "firstName", case_sensitive = false);
            const lastName = getProperty(borrower, "lastName", case_sensitive = false);
            const bn = _get_label_or_id(container);
            const date = getProperty(loan, "expectedReturn", case_sensitive = false);
            const subject = `Return Request for Container ${bn} accepted`;
            const body = `Dear ${firstName} ${lastName}, %0D%0D` +
                `please return container ${bn} on ${date}. %0D%0D` +
                `Kind Regards,%0D ${getUserName()}`;
            open_mail_generic(email, subject, body)
        } catch (err) {
            logger.error(err);
        }
    }


    var accept_return_function = async function (container) {
        const wait = getPleaseWaitSpan();
        const actions_panel = get_actions_panel(container);
        $(actions_panel).append(wait).find('btn').hide()

        const loan = (await get_active_loans(container))[0];

        const accept_return_request = {
            loan: getEntityID(loan),
        };

        const result = await run_script("loan_management/accept_return_request.py", accept_return_request);
        $(wait).remove();
        show_result(actions_panel, result, container);

        open_accept_return_mail(loan, container);
    }


    var reject_return_function = async function (container) {
        const wait = getPleaseWaitSpan();
        const actions_panel = get_actions_panel(container);
        $(actions_panel).append(wait).find('btn').hide()

        const loan = (await get_active_loans(container))[0];

        const reject_return_request = {
            loan: getEntityID(loan),
        };

        const result = await run_script(
            "loan_management/reject_return_request.py",
            reject_return_request);
        $(wait).remove();
        show_result(actions_panel, result, container);

        open_reject_return_mail(loan, container);
    }


    /**
     * Add two buttons for confirming or rejecting a return request.
     */
    var add_confirm_reject_return = function (container, loan) {
        const actionPanel = get_actions_panel(container);
        if (userIsAdministrator() === true || userHasRole("curator") === true) {
            const confirmButton = $(_accept_return_button);
            actionPanel.append(confirmButton);

            const rejectButton = $(_reject_return_button);
            actionPanel.append(rejectButton);

            confirmButton.click(() => {
                confirmButton.remove();
                rejectButton.remove();
                accept_return_function(container);
            });

            rejectButton.click(() => {
                confirmButton.remove();
                rejectButton.remove();
                reject_return_function(container);
            });
        } else {
            actionPanel.append($(create_link_to_loan(loan, "Pending return request.")));
        }
    }

    /**
     * Add a button for confirming the loan.
     */
    var add_confirm_loan_button = function (container, loan) {
        const actionsPanel = get_actions_panel(container);
        if (userIsAdministrator() === true || userHasRole("curator") === true) {
            const but = $(_confirm_loan_button);
            actionsPanel.append(but);

            but.click(() => {
                but.remove();
                confirm_loan_function(container);
            });
        } else {
            actionsPanel.append($(create_link_to_loan(loan, "Loan request accepted.")));
        }
    }

    var return_function = async function (form, container) {
        const wait = getPleaseWaitSpan();
        const actions_panel = get_actions_panel(container);
        $(actions_panel).append(wait).find('btn').hide()
        $(form).hide();

        const return_request = get_request_data(form);


        const loan = (await get_active_loans(container))[0];


        $.extend(return_request, {
            container: getEntityID(container),
            loan: getEntityID(loan),
        });

        const result = await run_script("loan_management/request_return.py", return_request);
        $(wait).remove();
        $(form).remove();
        show_result(actions_panel, result, container);
    }


    /**
     * Add buttons for borrowing containers.
     */
    var add_buttons = function (containers) {
        $(containers).find('.container-loan-btn').remove();
        $(containers).each(async function () {
            const loan_state = await get_loan_state(this);
            if (typeof loan_state.loan !== "undefined") {
                if (loan_state.state == "loan_requested") {
                    add_borrow_accept_button(this, loan_state.loan);
                } else if (loan_state.state == "loan_accepted") {
                    add_confirm_loan_button(this, loan_state.loan);
                } else if (loan_state.state == "lent") {
                    add_return_button(this);
                } else if (loan_state.state == "return_requested") {
                    add_confirm_reject_return(this, loan_state.loan);
                } else if (loan_state.state == "return_accepted") {
                    add_manual_return_button(this, loan_state.loan);
                }
            } else {
                add_borrow_button(this);
            }
        });
    }

    /**
     * This function generates a form for borrowing containers.
     * selectform can be used to generate the form for multiple containers.
     * In case selectform is true, container will be ignored.
     */
    var generate_form_borrow_checkout = async function (container) {
        const email = window.localStorage["borrower_email"];
        const firstname = window.localStorage["borrower_first_name"];
        const lastname = window.localStorage["borrower_last_name"];
        var exp_return = window.localStorage["borrower_return_date"];
        if (exp_return < Date.now()) {
            exp_return = "";
        }


        return getHTMLForm("loan-forms/borrow_checkout", {
            containerid: getEntityID(container),
            first_name: firstname,
            last_name: lastname,
            email: email,
            expected_return_date: exp_return,
            mindate: get_formatted_date("T")
        });
    }


    /**
     * Generate form for a new Return Request, with fields
     *
     * containerid (hidden)
     * first_name (pre-filled)
     * last_name (pre-filled)
     * email (pre-filled)
     * expectedreturn (pre-filled)
     * comment
     */
    var generate_return_form = async function (container) {
        const loan = (await get_active_loans(container))[0];
        const borrower = (await retrieve(getProperty(loan, "borrower", case_sensitive = false)))[0];

        var email = getProperty(borrower, "email", case_sensitive = false);
        var first_name = getProperty(borrower, "firstName",
            case_sensitive = false);
        var last_name = getProperty(borrower, "lastName",
            case_sensitive = false);
        var exp_return = getProperty(loan, "expectedReturn",
            case_sensitive = false);

        return getHTMLForm("loan-forms/return_container", {
            containerid: getEntityID(container),
            first_name: first_name,
            last_name: last_name,
            email: email,
            expectedreturn: exp_return,
            mindate: get_formatted_date("T")
        });
    }

    var init = function (containers) {
        const init_containers = containers || $(".caosdb-entity-panel")
            .not("[data-version-successor]")
            /* TODO: This fails easily (ContainerType) */
            .has(".caosdb-parent-name:contains('Container')")
            .not(":has(.caosdb-parent-name:contains('ContainerT'))").toArray();
        add_buttons(init_containers);
    }

    return {
        init: init,
        add_buttons: add_buttons,
        get_active_loans: get_active_loans,
        get_loan_state_string: get_loan_state_string,
    };

}();


/**
 * Add the extensions to the webui.
 */
$(document).ready(function () {
    // TODO build variable
    loan_management.init();
});
