#!/usr/bin/env python3
# encoding: utf-8
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#
"""
Accept a loan request.
"""
from __future__ import absolute_import
import caosdb as db
from caosadvancedtools.serverside.helper import print_success, get_timestamp
from container_loan import (main, get_loan, F_LOAN, assert_loan_state,
                            LOAN_ACCEPTED, S_LOAN_ACCEPTED, S_LOAN_REQUESTED,
                            get_borrower_names, set_property)


def _accept_loan_request(data):
    """Update a loan Record and add the `accepted` Property."""
    loan = get_loan(data[F_LOAN])
    assert_loan_state(loan, S_LOAN_REQUESTED)

    # This changes the state from "loan_requested" to "loan_accepted".
    set_property(loan, LOAN_ACCEPTED, get_timestamp())

    # To be sure that it worked:
    assert_loan_state(loan, S_LOAN_ACCEPTED)

    db.Container().extend([
        loan
    ]).update()

    return loan


def accept_loan_request(data):
    """Accept a loan request.

    I.e. update the `Loan` Record and add the `accepted` Property.
    """
    loan = _accept_loan_request(data)
    fn, ln = get_borrower_names(loan)

    print_success('Thank you for accepting the loan request by {fn} {ln}. See '
                  '<a href="{loan}" title="Go to the accepted loan '
                  'request.">here</a>'.format(fn=fn, ln=ln, loan=loan.id))
    return 0


if __name__ == "__main__":
    main(accept_loan_request)
