# encoding: utf-8
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#
"""
Creates a loan request using information provided by a web formular.
"""

from __future__ import absolute_import

import logging
import sys
import traceback
from datetime import datetime

import caosdb as db
from caosadvancedtools.serverside.helper import (DataModelError, get_data,
                                                 init_data_model,
                                                 parse_arguments, print_error,
                                                 print_info, print_warning,
                                                 send_mail)
from caosadvancedtools.serverside.logging import configure_server_side_logging
from caosdb.exceptions import EmptyUniqueQueryError
from validate_email import validate_email

LOGGER_NAME = "container_loan"
LOGGER = logging.getLogger(LOGGER_NAME)

# Form names
F_CONTAINER = "container"
F_FIRST_NAME = "first_name"
F_LAST_NAME = "last_name"
F_EMAIL = "email"
F_EXPECTED_RETURN_DATE = "expected_return_date"
F_EXHAUST_CONTENTS = "exhaust_contents"
F_COMMENT = "comment"
F_LOAN = "loan"

S_RETURNED = "returned"
S_RETURN_ACCEPTED = "return_accepted"
S_RETURN_REQUESTED = "return_requested"
S_LENT = "lent"
S_LOAN_ACCEPTED = "loan_accepted"
S_LOAN_REQUESTED = "loan_requested"


_EMTPY_DATA = {
    F_CONTAINER: None,
    F_FIRST_NAME: None,
    F_LAST_NAME: None,
    F_EMAIL: None,
    F_EXPECTED_RETURN_DATE: None,
    F_EXHAUST_CONTENTS: None,
    F_COMMENT: None,
    F_LOAN: None,
}

CONTAINER = db.RecordType(name="Container")
CONTAINER_RETURNED = "Container (returned)"
CONTAINER_BORROWED = "Container (borrowed)"
PERSON = db.RecordType(name="Person")
LOAN = db.RecordType(name="Loan")

RECORD_TYPES = [
    CONTAINER,
    PERSON,
    LOAN,
]

# Properties
FIRST_NAME = db.Property(name="firstName", datatype=db.TEXT)
LAST_NAME = db.Property(name="lastName", datatype=db.TEXT)
EMAIL = db.Property(name="email", datatype=db.TEXT)
COMMENT = db.Property(name="comment", datatype=db.TEXT)
EXHAUST_CONTENTS = db.Property(name="exhaustContents", datatype=db.BOOLEAN)
BORROWER = db.Property(name="Borrower", datatype=PERSON.name)
CONTENT = db.Property(name="Content", datatype=db.TEXT)
LOAN_REQUESTED = db.Property(name="loanRequested", datatype=db.DATETIME)
EXPECTED_RETURN = db.Property(name="expectedReturn", datatype=db.DATETIME)
LOAN_ACCEPTED = db.Property(name="loanAccepted", datatype=db.DATETIME)
LENT = db.Property(name="lent", datatype=db.DATETIME)
RETURN_REQUESTED = db.Property(name="returnRequested", datatype=db.DATETIME)
RETURN_ACCEPTED = db.Property(name="returnAccepted", datatype=db.DATETIME)
RETURNED = db.Property(name="returned", datatype=db.DATETIME)

PROPERTIES = [
    FIRST_NAME,
    LAST_NAME,
    EMAIL,
    COMMENT,
    EXHAUST_CONTENTS,
    BORROWER,
    CONTENT,
    LOAN_REQUESTED,
    EXPECTED_RETURN,
    LOAN_ACCEPTED,
    LENT,
    RETURN_REQUESTED,
    RETURN_ACCEPTED,
    RETURNED,
]


def _get_mail_from():
    config = db.get_config()
    result = "admin@indiscale.com"

    if "Misc" in config and "entity_loan.curator_mail_from" in config["Misc"]:
        result = config["Misc"]["entity_loan.curator_mail_from"]

    return result


def _get_mail_curator():
    config = db.get_config()
    result = "admin@indiscale.com"

    if "Misc" in config and "entity_loan.curator_mail_to" in config["Misc"]:
        result = config["Misc"]["entity_loan.curator_mail_to"]

    return result


EMAIL_FROM = _get_mail_from()
EMAIL_CURATOR = _get_mail_curator()


class MissingParentError(RuntimeError):

    def __init__(self, entity, expected_parent):
        super().__init__(
            "This entity should have a {par} parent:\n{ent}".format(
                par=expected_parent.name,
                ent=str(entity)))


class StateError(RuntimeError):

    def __init__(self, actual, expected):
        one = "" if len(expected) == 1 else "one of "
        _expected = expected[0] if len(expected) == 1 else ",".join(
            expected[:-1]) + ", or" + expected[-1]
        super().__init__(
            "This transition of the loan state is not possible. The expected "
            "state is {one}'{expected}' "
            "while the actual state is '{actual}'.".format(one=one,
                                                           actual=actual,
                                                           expected=_expected))


class EmailPatternError(RuntimeError):

    def __init__(self, email):
        super().__init__("The email address '{}' is not valid.".format(email))


class DataError(RuntimeError):

    def __init__(self, msg):
        super().__init__("The request most likely contained a data error."
                         " {}Please check your input.".format(msg if msg is not None else ""))


def assert_key_in_data(data, key):
    if key not in data:
        raise DataError("Missing information. "
                        "The data is expected to contain a `{}` "
                        "property.".format(key))


def assert_is_date(string, msg=None):
    try:
        return datetime.fromisoformat(string)
    except (TypeError, ValueError):
        pass  # and raise another error

    _msg = msg if msg is not None else ("Should be a ISO 8601 date "
                                        "string. "
                                        "Was '{}'. ".format(string))
    raise DataError(_msg)


def assert_date_in_future(actual, msg=None):
    date = assert_is_date(actual)

    if date <= datetime.now(tz=date.tzinfo):
        _msg = msg if msg is not None else ("Date is to be in the future. "
                                            "Was '{}'. ".format(actual))
        raise DataError(_msg)


def get_record_by_id(parent, entity_id):
    """ Retrieve a record with a particular parent by id. """

    return db.execute_query("FIND RECORD {} WITH ID = {}".format(parent,
                                                                 entity_id),
                            unique=True)


def get_container(container):
    """ Retrieve a container record by id. """

    return get_record_by_id(CONTAINER.name, container)


def get_loan(loan):
    """ Retrieve a loan record by id. """

    return get_record_by_id(LOAN.name, loan)


def get_loan_state(loan):
    if loan.get_parent(LOAN) is None:
        raise MissingParentError(loan, LOAN)

    if loan.get_property(RETURNED.name) is not None:
        return S_RETURNED

    if loan.get_property(RETURN_ACCEPTED.name) is not None:
        return S_RETURN_ACCEPTED

    if loan.get_property(RETURN_REQUESTED.name) is not None:
        return S_RETURN_REQUESTED

    if loan.get_property(LENT.name) is not None:
        return S_LENT

    if loan.get_property(LOAN_ACCEPTED.name) is not None:
        return S_LOAN_ACCEPTED

    if loan.get_property(LOAN_REQUESTED.name) is not None:
        return S_LOAN_REQUESTED


def assert_loan_state(loan, state):
    actual = get_loan_state(loan)
    expected = state if isinstance(state, list) else [state]

    if actual not in expected:
        raise StateError(actual, expected)


def get_borrower_names(loan):
    """Retrieve the borrower's first name and last name.

    Intented for generating a nice and informative success message.

    If the loan has no borrower ["an", "unknown borrower"] is returned.

    Parameters
    ----------
    loan : db.Record
        A `loan` Record which should have a `borrower` Property.
    """
    borrower_property = loan.get_property(BORROWER.name)

    if borrower_property is None:
        # The loan had no borrower. This should not happen, but we don't want
        # it to break the functionality.
        # TODO issue warning
        fn = "an"
        ln = "unknown borrower"
    else:
        borrower = db.Record(id=borrower_property.value).retrieve()
        fn = borrower.get_property(FIRST_NAME.name).value
        ln = borrower.get_property(LAST_NAME.name).value

    return fn, ln


def assert_email_pattern(email):
    """Raise an EmailPatternError if the email does't match.

    Parameters
    ----------
    email: str

    Raises
    ------
    EmailPatternError
    """

    if email is None or not validate_email(email_address=email,
                                           check_format=True, check_smtp=False,
                                           check_dns=False,
                                           check_blacklist=False):
        raise EmailPatternError(email)


def get_property_value(entity, prop, fallback=None):
    p = entity.get_property(prop)

    if p is None or p.value is None:
        return fallback

    return p.value


def get_requester_string(requester):
    results = [
        get_property_value(requester, FIRST_NAME, fallback=""),
        get_property_value(requester, LAST_NAME, fallback=""),
        get_property_value(requester, EMAIL, fallback="")
    ]

    # wrap email in < and >

    if len(results[2]) > 0:
        results[2] = "<" + results[2] + ">"

    return " ".join(results)


def get_external_server_uri():
    uri = db.get_config().get("sss_helper", "external_uri")

    if uri is not None:
        if not uri.endswith("/"):
            uri = uri + "/"

        return uri
    raise RuntimeError("Could not determine the external server uri")


def send_loan_request_mail(data, borrower, loan):
    try:
        containerid = str(loan.get_property(CONTAINER).value).split("@")[0]
        container = get_container(containerid)

        link = "{uri}Entity/{containerid}&{loanid}".format(
            uri=get_external_server_uri(),
            containerid=containerid,
            loanid=loan.id)

        body = """Dear Curator,

    a new loan has been request by {borrower} for container {containername}.

    Loan request:
        {data}

    View container and loan records:
        {link}
    """.format(containername=container.name, borrower=get_requester_string(borrower),
               data=data, link=link)

        send_mail(
            from_addr=EMAIL_FROM,
            to=EMAIL_CURATOR,
            subject="loan request",
            body=body)
        print_info("An email has been sent to the responsible container curator.")
    except BaseException as e:
        print_error("""Sending an email to the responsible container curator failed

for an unknown reason. Please inform the container curator by yourself about your
loan request and that the email sending failed. Thank you very much for your
help.""")
        LOGGER.error(e)


def send_return_request_mail(data, returner, loan):
    try:
        containerid = str(loan.get_property(CONTAINER).value).split("@")[0]
        container = get_container(containerid)

        link = "{uri}Entity/{containerid}&{loanid}".format(
            uri=get_external_server_uri(),
            containerid=containerid,
            loanid=loan.id)

        body = """Dear Curator,

    a new return request by {returner} for container {containername} is pending.

    Return request:
        {data}

    View container and loan records:
        {link}
    """.format(containername=container.name, returner=get_requester_string(returner),
               data=data, link=link)

        send_mail(
            from_addr=EMAIL_FROM,
            to=EMAIL_CURATOR,
            subject="return request",
            body=body)
        print_info("An email has been sent to the responsible container curator.")
    except BaseException as e:
        print_error("""Sending an email to the responsible container curator failed

for an unknown reason. Please inform the container curator by yourself about this

return request and that the email sending failed. Thank you very much for your
help.""")
        LOGGER.error(e)


def create_person(firstname, lastname, email):
    """ Insert a new person Record."""
    person = db.Record().add_parent(PERSON)
    person.add_property(FIRST_NAME, firstname)
    person.add_property(LAST_NAME, lastname)
    assert_email_pattern(email)
    person.add_property(EMAIL, email)

    return person


def query_person(firstname, lastname, email):
    """ Query a person

    The person is identified by
    1. the email address
    2. or the first name and the last name
    """
    query = ('FIND RECORD {rt_person} '
             'WITH {pemail} = "{email}" '
             'OR ({pfn} = "{firstname}" '
             'AND {pln} = "{lastname}") '
             ).format(
        rt_person=PERSON.name,
        pemail=EMAIL.name,
        email=email,
        pfn=FIRST_NAME.name,
        firstname=firstname,
        pln=LAST_NAME.name,
        lastname=lastname
    )

    return db.execute_query(query, unique=True)


def _update_person(person, firstname, lastname, email):
    """Update the persons names and email address.

    But only if anything changes. That said, if only the case of the names or
    the email address changes, the person entity is left unchanged.
    """

    if set_property(
            person, EMAIL, email, False) or set_property(
            person, FIRST_NAME, firstname, False) or set_property(
            person, LAST_NAME, lastname, False):
        person.update()


def get_person(firstname, lastname, email):
    """ Retrieve a person Record or create it. """
    try:
        person = query_person(firstname, lastname, email)

        # TODO this is a side effect which might not be desired in all
        # cases. How should we handle it?
        _update_person(person, firstname, lastname, email)

        return person
    except EmptyUniqueQueryError:
        return create_person(firstname, lastname, email)


def set_property(entity, prop, value, case_sensitive=True):
    """Set a property value.

    If the property already has that value, the value is left untouched.

    Parameters
    ----------
    entity : db.Entity
    prop : str or db.Property
    value : object
    case_sensitive : boolean (optional)
        defaults to 'True'. Only matters if `value` is a string.
    """
    p = entity.get_property(prop)

    if p is None and value is not None:
        entity.add_property(prop, value)

        return True
    elif ((case_sensitive is True and p.value != value)
          or (hasattr(p.value, "lower")
              and hasattr(value, "lower")
              and p.value.lower() != value.lower())):
        p.value = value

        return True

    return False


def check_data_model():
    return init_data_model(RECORD_TYPES + PROPERTIES)


def _caller(func, args):
    params = parse_arguments(args)

    if params.auth_token:
        db.configure_connection(password_method="auth_token",
                                auth_token=params.auth_token)
        check_data_model()
    else:
        print_warning("No auth_token passed.")

    data = get_data(params.filename, default=_EMTPY_DATA)
    LOGGER.info(
        "handling entity loan request: {args}, {data}".format(args=args,
                                                              data=data))

    return func(data)


def main(func, args=None):
    # configure_server_side_logging(LOGGER_NAME)
    try:
        code = _caller(func, args if args is not None else sys.argv[1:])
    except SystemExit:
        raise
    except DataModelError as e:
        print_error(e.args[0])
        code = 3
        LOGGER.debug(e)
    except DataError as e:
        print_error(e.args[0])
        code = 4
        LOGGER.debug(e)
    except BaseException as e:  # pylint: disable=W0703
        tb = traceback.format_exc()
        print_error("An unknown error occured."
                    "<br>{}<code>{}</code>".format(str(e), tb))
        code = 1
        LOGGER.debug("Unknown error!!!", e)
    LOGGER.info("exiting with code {}".format(code))
    sys.exit(code)
