#!/usr/bin/env python3
# encoding: utf-8
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#
"""
Confirm a loan.
"""
from __future__ import absolute_import

import caosdb as db
from caosadvancedtools.serverside.helper import get_timestamp, print_success

from container_loan import (CONTAINER, CONTAINER_BORROWED, F_LOAN, LENT, S_LENT,
                            S_LOAN_ACCEPTED, assert_loan_state, get_borrower_names,
                            get_loan, main, set_property)


def _set_lent_container(loan):
    """Set the container version to HEAD.

    This stores the version of the container when it was delivered to the borrower.
    """
    container_prop = loan.get_property(CONTAINER)
    container_prop.name = CONTAINER_BORROWED
    container_prop.value = str(container_prop.value) + "@HEAD"


def _confirm_loan(data):
    """Update a loan Record and add the `lent` Property."""
    loan = get_loan(data[F_LOAN])
    assert_loan_state(loan, S_LOAN_ACCEPTED)

    # This changes the state from "loan_accepted" to "lent".
    set_property(loan, LENT, get_timestamp())
    _set_lent_container(loan)

    # To be sure that it worked:
    assert_loan_state(loan, S_LENT)

    db.Container().extend([
        loan
    ]).update()

    return loan


def confirm_loan(data):
    """Confirm a loan.

    I.e. update the `Loan` Record and add the `lent` Property.
    """
    loan = _confirm_loan(data)
    fn, ln = get_borrower_names(loan)
    loan = get_loan(data[F_LOAN])
    container_id = loan.get_property(CONTAINER_BORROWED).value.split("@")[0]
    print_success('Thank you for confirming the loan request by {fn} {ln}.<br>'
                  'You can also checkout the new loan state '
                  '<a href="{loan}" title="Go to the confirmed loan '
                  'record.">here</a>'.format(fn=fn, ln=ln, loan=loan.id))

    return 0


if __name__ == "__main__":
    main(confirm_loan)
