#!/usr/bin/env python3
# encoding: utf-8
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#
"""
Manually return a container.
"""
from __future__ import absolute_import

import caosdb as db
from caosadvancedtools.serverside.helper import get_timestamp, print_success

from container_loan import (CONTAINER, CONTAINER_BORROWED, CONTAINER_RETURNED,
                            CONTENT, F_LOAN, RETURNED, S_RETURN_ACCEPTED,
                            S_RETURNED, assert_loan_state, get_borrower_names,
                            get_loan, main, set_property)


def _set_returned_container(loan):
    """Add a `Container (returned)` property to the loan.

    This stores the version of the container that was returned by a borrower.
    """
    container_id = loan.get_property(CONTAINER_BORROWED).value.split("@")[0]
    loan.add_property(property=CONTAINER,
                      name=CONTAINER_RETURNED,
                      value=container_id + "@HEAD")


def set_content(loan):
    """Set the content to the container to the one given in the loan.  """
    container_id = loan.get_property(CONTAINER_BORROWED).value.split("@")[0]
    container = db.Record(id=container_id).retrieve()
    set_property(container, CONTENT, loan.get_property(CONTENT).value)
    container.update()


def _manual_return(data):
    """Update a loan Record and add the `returned` Property."""
    loan = get_loan(data[F_LOAN])
    assert_loan_state(loan, S_RETURN_ACCEPTED)

    # This changes the state from "return_accepted" to "returned".
    set_property(loan, RETURNED, get_timestamp())
    set_property(loan, RETURNED, get_timestamp())

    # To be sure that it worked:
    assert_loan_state(loan, S_RETURNED)

    # add returned container
    _set_returned_container(loan)

    db.Container().extend([
        loan
    ]).update()

    return loan


def manual_return(data):
    """Return a container from a borrower.

    I.e. update the `Loan` Record and add the `returned` Property.
    """
    loan = _manual_return(data)
    fn, ln = get_borrower_names(loan)
    loan = get_loan(data[F_LOAN])

    print_success('The container borrowed by {fn} {ln} '.format(fn=fn, ln=ln)
                  + 'has been returned.<br>')

    return 0


if __name__ == "__main__":
    main(manual_return)
