#!/usr/bin/env python3
# encoding: utf-8
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#
"""
Creates a return request using information provided by a web formular.

A return request is a loan record with a returnRequested property.
"""


from __future__ import absolute_import

from caosadvancedtools.serverside.helper import get_timestamp, print_success

from container_loan import (BORROWER, COMMENT, CONTENT, EXPECTED_RETURN, F_COMMENT,
                            F_EMAIL, F_EXPECTED_RETURN_DATE,
                            F_FIRST_NAME, F_LAST_NAME, F_LOAN, FIRST_NAME, LAST_NAME,
                            RETURN_REQUESTED, S_LENT,
                            assert_date_in_future, assert_key_in_data,
                            assert_loan_state, get_loan, get_person, main,
                            send_return_request_mail, set_property)


def _check_data(data):
    assert_key_in_data(data, F_EXPECTED_RETURN_DATE)
    assert_key_in_data(data, F_LOAN)
    assert_key_in_data(data, F_COMMENT)

    assert_date_in_future(
        data[F_EXPECTED_RETURN_DATE],
        ("The expected return date needs to be in the future."
         " You submitted '{}'.").format(
            data[F_EXPECTED_RETURN_DATE]))


def _issue_return_request(data):
    _check_data(data)
    loan = get_loan(data[F_LOAN])
    assert_loan_state(loan, S_LENT)

    returner = get_person(data[F_FIRST_NAME], data[F_LAST_NAME], data[F_EMAIL])
    loan.add_property(RETURN_REQUESTED, get_timestamp())

    set_property(loan, BORROWER, returner.id)
    set_property(loan, EXPECTED_RETURN, data[F_EXPECTED_RETURN_DATE])
    set_property(loan, CONTENT, data[F_COMMENT])

    loan.update()

    return returner, loan


def issue_return_request(data):
    """ Issue a return request.

    A return request is a Loan record where a returnRequested property is being
    added.
    """
    returner, loan = _issue_return_request(data)

    # print
    fn = returner.get_property(FIRST_NAME.name).value
    ln = returner.get_property(LAST_NAME.name).value
    print_success(
        'Thank you for your return request, {fn} {ln}.'.format(
            fn=fn, ln=ln))

    # send an email and inform the curator about the return request
    send_return_request_mail(data, returner, loan)

    return 0


if __name__ == "__main__":
    main(issue_return_request)
